#include "php.h"
#include "zend_compile.h"
#include "php_ini.h"
#include "php_globals.h"
#include "php_orbit.h"
#include "phorb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define ALIGN_VAL(buf, align)		*buf = ALIGN_ADDRESS(*buf, align)
#define GET_VAL(buf, type)		(*(type *)(*buf))
#define ADVANCE_PTR(buf, offset)	*buf = ((guchar *)(*buf)) + offset

zval *phorb_get_zval(gconstpointer *buf, CORBA_TypeCode tc);

static zval *get_null(gconstpointer *buf, zval *res)
{
	ZVAL_NULL(res);

	return res;
}

static zval *get_short(gconstpointer *buf, zval *res)
{
	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_SHORT);
	ZVAL_LONG(res, GET_VAL(buf, CORBA_short));
	ADVANCE_PTR(buf, sizeof(CORBA_short));

	return res;
}

static zval *get_ushort(gconstpointer *buf, zval *res) {
	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_SHORT);
	ZVAL_LONG(res, GET_VAL(buf, CORBA_unsigned_short));
	ADVANCE_PTR(buf, sizeof(CORBA_unsigned_short));

	return res;
}

static zval *get_long(gconstpointer *buf, zval *res)
{
	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG);
	ZVAL_LONG(res, GET_VAL(buf, CORBA_long));
	ADVANCE_PTR(buf, sizeof(CORBA_long));

	return res;
}

static zval *get_ulong(gconstpointer *buf, zval *res)
{
	CORBA_unsigned_long v;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG);
	v = GET_VAL(buf, CORBA_unsigned_long);
	ADVANCE_PTR(buf, sizeof v);

	if ((unsigned long)v > LONG_MAX) {
		ZVAL_DOUBLE(res, (double)v);
	} else {
		ZVAL_LONG(res, (long)v);
	}

	return res;
}

static zval *get_longlong(gconstpointer *buf, zval *res)
{
	CORBA_long_long v;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG_LONG);
	v = GET_VAL(buf, CORBA_long_long);
	ADVANCE_PTR(buf, sizeof v);

	if (v < LONG_MIN || v > LONG_MAX) {
		ZVAL_DOUBLE(res, (double)v);
	} else {
		ZVAL_LONG(res, (long)v);
	}

	return res;
}

static zval *get_ulonglong(gconstpointer *buf, zval *res)
{
	CORBA_unsigned_long_long v;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG_LONG);
	v = GET_VAL(buf, CORBA_unsigned_long_long);
	ADVANCE_PTR(buf, sizeof v);

	if (v > LONG_MAX) {
		ZVAL_DOUBLE(res, (double)v);
	} else {
		ZVAL_LONG(res, (long)v);
	}

	return res;
}

static zval *get_float(gconstpointer *buf, zval *res)
{
	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_FLOAT);
	ZVAL_DOUBLE(res, GET_VAL(buf, CORBA_float));
	ADVANCE_PTR(buf, sizeof(CORBA_float));

	return res;
}

static zval *get_double(gconstpointer *buf, zval *res)
{
	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_DOUBLE);
	ZVAL_DOUBLE(res, GET_VAL(buf, CORBA_double));
	ADVANCE_PTR(buf, sizeof(CORBA_double));

	return res;
}

static zval *get_longdouble(gconstpointer *buf, zval *res) {
	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG_DOUBLE);
	ZVAL_DOUBLE(res, GET_VAL(buf, CORBA_long_double));
	ADVANCE_PTR(buf, sizeof(CORBA_long_double));

	return res;
}

static zval *get_boolean(gconstpointer *buf, zval *res) {
	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_BOOLEAN);
	ZVAL_BOOL(res, GET_VAL(buf, CORBA_boolean));
	ADVANCE_PTR(buf, sizeof(CORBA_boolean));

	return res;
}

static zval *get_char(gconstpointer *buf, zval *res) {
	CORBA_char tmp[2];

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_CHAR);
	tmp[0] = GET_VAL(buf, CORBA_char);
	tmp[1] = '\0';
	ZVAL_STRING(res, tmp, 1);
	ADVANCE_PTR(buf, sizeof(CORBA_char));

	return res;
}

static zval *get_wchar(gconstpointer *buf, zval *res) {
	CORBA_wchar tmp[2];
	GError *error = NULL;
	gchar *utf8;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_WCHAR);
	tmp[0] = GET_VAL(buf, CORBA_wchar);
	tmp[1] = '\0';

	utf8 = g_utf16_to_utf8(tmp, -1, NULL, NULL, &error);
	if (!utf8) {
		zend_error(E_WARNING, "UTF16 read error: %s", error->message);
		g_error_free(error);
		return NULL;
	}
	ZVAL_STRING(res, utf8, 1);
	g_free(utf8);

	ADVANCE_PTR(buf, sizeof(CORBA_wchar));

	return res;
}

static zval *get_octet(gconstpointer *buf, zval *res) {
	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_OCTET);
	ZVAL_LONG(res, GET_VAL(buf, CORBA_octet));
	ADVANCE_PTR(buf, sizeof(CORBA_octet));

	return res;
}

static zval *get_enum(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	CORBA_unsigned_long v;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_LONG);
	v = GET_VAL(buf, CORBA_long);
	ADVANCE_PTR(buf, sizeof v);
	if (v > tc->sub_parts) {
		zend_error(E_WARNING,
		           "enumeration received with invalid value (%lu)", v);
		return NULL;
	}
	ZVAL_STRING(res, tc->subnames[v], 1);

	return res;
}

static zval *get_struct(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	CORBA_unsigned_long i;

	array_init(res);
	for(i = 0; i < tc->sub_parts; i++) {
		zval *val = phorb_get_zval(buf, tc->subtypes[i]);

		if (!val)
			return NULL;

		zend_hash_update(HASH_OF(res), (char *)tc->subnames[i],
		                 strlen(tc->subnames[i]) + 1,
		                 (void *)&val, sizeof(zval *), NULL);
	}
	convert_to_object(res);

	return res;
}

static zval *get_sequence(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	const CORBA_sequence_CORBA_octet *sval;
	gconstpointer seqval;
	int i, align;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_SEQ);
	sval = (CORBA_sequence_CORBA_octet *)*buf;
	align = tc->subtypes[0]->c_align;
	seqval = sval->_buffer;
	ADVANCE_PTR(buf, sizeof(CORBA_sequence_CORBA_octet));

	array_init(res);
	for(i = 0; i < sval->_length; i++) {
		zval *val;

		ALIGN_VAL(&seqval, align);
		val = phorb_get_zval(&seqval, tc->subtypes[0]);
		if (!val)
			return NULL;
		zend_hash_next_index_insert(HASH_OF(res), (void *)&val,
		                           sizeof(zval *), NULL);
	}

	return res;
}

static zval *get_array(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	int i, align = tc->subtypes[0]->c_align;

	array_init(res);
	for(i = 0; i < tc->length; i++) {
		zval *val;

		ALIGN_VAL(buf, align);
		val = phorb_get_zval(buf, tc->subtypes[0]);
		if (!val)
			return NULL;
		zend_hash_next_index_insert(HASH_OF(res), (void *)&val,
		                           sizeof(zval *), NULL);
	}

	return res;
}

static zval *get_objref(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	CORBA_Object obj;
	zval *handle;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_POINTER);
	obj = GET_VAL(buf, CORBA_Object);
	ADVANCE_PTR(buf, sizeof obj);

	if (!obj)
		return get_null(buf, res);

	res = phorb_init_object(res);
	if (!res)
		return NULL;

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_object(
	                     	CORBA_Object_duplicate(obj, NULL)));
	zend_hash_index_update(Z_OBJPROP_P(res), 0, &handle,
	                       sizeof(zval *), NULL);

	return res;
}

static zval *get_string(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	CORBA_unsigned_long len;
	CORBA_char *str;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_POINTER);
	str = GET_VAL(buf, CORBA_string);
	ADVANCE_PTR(buf, sizeof(CORBA_string));

	if (str) {
		ZVAL_STRING(res, str, 1);
	} else {
		return get_null(buf, res);
	}

	return res;
}

static zval *get_wstring(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	CORBA_unsigned_long len;
	GError *error = NULL;
	CORBA_wchar *wstr;
	gchar *utf8;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_POINTER);
	wstr = GET_VAL(buf, CORBA_wstring);
	ADVANCE_PTR(buf, sizeof(CORBA_string));
	if (!wstr)
		return get_null(buf, res);

	utf8 = g_utf16_to_utf8(wstr, -1, NULL, NULL, &error);
	if (!utf8) {
		zend_error(E_WARNING, "UTF16 read error: %s", error->message);
		g_error_free(error);
		return NULL;
	}
	ZVAL_STRING(res, utf8, 1);
	g_free(utf8);

	return res;
}

static zval *get_fixed(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	return NULL;	/* FIXME: implement this */
}

static zval *get_union(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	gconstpointer tmp;
	CORBA_long arm;
	zval *discriminator, *value;
	int i, sz = 0;

	ALIGN_VAL(buf, MAX(tc->c_align, tc->discriminator->c_align));
	discriminator = phorb_get_zval(buf, tc->discriminator);
	if (!discriminator)
		return NULL;

	arm = phorb_union_find_arm(tc, discriminator);
	if (arm < 0) {
		zval_dtor(discriminator);
		FREE_ZVAL(discriminator);

		return get_null(buf, res);
	}

	for(i = 0; i < tc->sub_parts; i++)
		sz = MAX(sz, ORBit_gather_alloc_info(tc->subtypes[i]));

	ALIGN_VAL(buf, tc->c_align);
	tmp = *buf;
	ADVANCE_PTR(buf, sz);

	value = phorb_get_zval(&tmp, tc->subtypes[arm]);

	array_init(res);
	zend_hash_next_index_insert(HASH_OF(res), (void *)&discriminator,
	                            sizeof(zval *), NULL);
	zend_hash_next_index_insert(HASH_OF(res), (void *)&value,
	                            sizeof(zval *), NULL);

	return res;
}

static zval *get_typecode(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	CORBA_TypeCode obj;
	zval *handle;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_POINTER);
	obj = GET_VAL(buf, CORBA_TypeCode);
	ADVANCE_PTR(buf, sizeof obj);

	if (!obj)
		return get_null(buf, res);

	res = phorb_init_typecode(res);
	if (!res)
		return NULL;

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_typecode(
	                     	(CORBA_TypeCode)CORBA_Object_duplicate(
	                 		(CORBA_Object)obj, NULL)));
	zend_hash_index_update(Z_OBJPROP_P(res), 0, &handle,
	                       sizeof(zval *), NULL);

	return res;
}

static zval *get_any(gconstpointer *buf, CORBA_TypeCode tc, zval *res)
{
	gconstpointer anyval;
	CORBA_any *any;
	zval *valtc, *value, *handle, *tmp;

	ALIGN_VAL(buf, ORBIT_ALIGNOF_CORBA_ANY);
	any = &GET_VAL(buf, CORBA_any);
	ADVANCE_PTR(buf, sizeof(CORBA_any));

	MAKE_STD_ZVAL(valtc);
	tmp = phorb_init_typecode(valtc);
	if (!tmp) {
		FREE_ZVAL(valtc);
		return NULL;
	} else
		valtc = tmp;

	MAKE_STD_ZVAL(handle);
	ZVAL_RESOURCE(handle, phorb_register_typecode(
	                     	(CORBA_TypeCode)CORBA_Object_duplicate(
	                 		(CORBA_Object)any->_type, NULL)));
	zend_hash_index_update(Z_OBJPROP_P(valtc), 0, &handle,
	                       sizeof(zval *), NULL);

	anyval = any->_value;
	value = phorb_get_zval(&anyval, any->_type);
	if (!value) {
		FREE_ZVAL(handle);
		FREE_ZVAL(valtc);
		return NULL;
	}

	array_init(res);
	zend_hash_next_index_insert(HASH_OF(res), (void *)&valtc,
	                            sizeof(zval *), NULL);
	zend_hash_next_index_insert(HASH_OF(res), (void *)&value,
	                            sizeof(zval *), NULL);

	return res;
}

zval *phorb_get_zval(gconstpointer *buf, CORBA_TypeCode tc)
{
	zval *res, *ret = NULL;

	while(tc->kind == CORBA_tk_alias)
		tc = tc->subtypes[0];

	MAKE_STD_ZVAL(res);
	ZVAL_NULL(res);

	switch(tc->kind) {
	case CORBA_tk_null:
		ret = get_null(buf, res);
		break;
	case CORBA_tk_void:
		ret = NULL;
		break;
	case CORBA_tk_short:
		ret = get_short(buf, res);
		break;
	case CORBA_tk_ushort:
		ret = get_ushort(buf, res);
		break;
	case CORBA_tk_long:
		ret = get_long(buf, res);
		break;
	case CORBA_tk_ulong:
		ret = get_ulong(buf, res);
		break;
	case CORBA_tk_longlong:
		ret = get_longlong(buf, res);
		break;
	case CORBA_tk_ulonglong:
		ret = get_ulonglong(buf, res);
		break;
	case CORBA_tk_float:
		ret = get_float(buf, res);
		break;
	case CORBA_tk_double:
		ret = get_double(buf, res);
		break;
	case CORBA_tk_longdouble:
		ret = get_longdouble(buf, res);
		break;
	case CORBA_tk_char:
		ret = get_char(buf, res);
		break;
	case CORBA_tk_wchar:
		ret = get_wchar(buf, res);
		break;
	case CORBA_tk_boolean:
		ret = get_boolean(buf, res);
		break;
	case CORBA_tk_octet:
		ret = get_octet(buf, res);
		break;
	case CORBA_tk_except:
		/* exceptions are demarshalled like structures */
	case CORBA_tk_struct:
		ret = get_struct(buf, tc, res);
		break;
	case CORBA_tk_objref:
		ret = get_objref(buf, tc, res);
		break;
	case CORBA_tk_enum:
		ret = get_enum(buf, tc, res);
		break;
	case CORBA_tk_sequence:
		ret = get_sequence(buf, tc, res);
		break;
	case CORBA_tk_union:
		ret = get_union(buf, tc, res);
		break;
	case CORBA_tk_string:
		ret = get_string(buf, tc, res);
		break;
	case CORBA_tk_wstring:
		ret = get_wstring(buf, tc, res);
		break;
	case CORBA_tk_array:
		ret = get_array(buf, tc, res);
		break;
	case CORBA_tk_TypeCode:
		ret = get_typecode(buf, tc, res);
		break;
	case CORBA_tk_any:
		ret = get_any(buf, tc, res);
		break;
	case CORBA_tk_fixed:
		ret = get_fixed(buf, tc, res);
		break;
	case CORBA_tk_Principal:
	default:
		zend_error(E_WARNING,
		           "Unsupported input typecode: `%s' (%d)\n",
		           tc->repo_id, tc->kind);
	}

	if (!ret) {
		zval_dtor(res);
		FREE_ZVAL(res);
		return NULL;
	}

	return ret;
}
